from __future__ import annotations

import asyncio
import importlib
import json
import os
import shlex
import uuid
from typing import Any, Dict, List, Optional, Tuple


class PythonSDKMCPClient:
    """
    Optional MCP transport using the official Python SDK over stdio.
    This is only activated when PAPERFIG_MCP_COMMAND is configured.
    """

    def __init__(self, command: str, args: Optional[List[str]] = None) -> None:
        self.command = command
        self.args = args or []

    def call_tool(self, server: Optional[str], tool: str, arguments: Dict[str, Any]) -> Dict[str, Any]:
        # `server` is ignored for direct stdio transport; included for API compatibility.
        del server
        return asyncio.run(self._call_tool(tool=tool, arguments=arguments))

    async def _call_tool(self, tool: str, arguments: Dict[str, Any]) -> Dict[str, Any]:
        try:
            from mcp import ClientSession, StdioServerParameters  # type: ignore
            from mcp.client.stdio import stdio_client  # type: ignore
        except Exception as exc:  # pragma: no cover - optional dependency
            raise RuntimeError(
                "MCP SDK not installed. Install optional dependency `mcp`."
            ) from exc

        params = StdioServerParameters(command=self.command, args=self.args)
        async with stdio_client(params) as (read, write):
            async with ClientSession(read, write) as session:
                await session.initialize()
                result = await session.call_tool(tool, arguments)
        return _normalize_mcp_response(result)


def _normalize_mcp_response(result: Any) -> Dict[str, Any]:
    if isinstance(result, dict):
        payload = result
    elif hasattr(result, "model_dump"):
        payload = result.model_dump()  # type: ignore[assignment]
    elif hasattr(result, "dict"):
        payload = result.dict()  # type: ignore[assignment]
    else:
        payload = {}

    if isinstance(payload.get("svg"), str):
        return payload

    for container_key in ("result", "data", "output"):
        candidate = payload.get(container_key)
        if isinstance(candidate, dict) and isinstance(candidate.get("svg"), str):
            return candidate

    structured = payload.get("structuredContent")
    if isinstance(structured, dict):
        return structured

    text_chunks: List[str] = []
    content_items = payload.get("content", [])
    for item in content_items:
        if isinstance(item, dict):
            text_value = item.get("text")
        else:
            text_value = getattr(item, "text", None)
        if isinstance(text_value, str):
            text_chunks.append(text_value)

    for chunk in text_chunks:
        try:
            parsed = json.loads(chunk)
        except json.JSONDecodeError:
            continue
        if isinstance(parsed, dict):
            return parsed

    raise RuntimeError("MCP tool response did not include parseable structured content.")


def _load_client_factory(factory_spec: str) -> Any:
    if ":" not in factory_spec:
        raise RuntimeError(
            "Invalid PAPERFIG_MCP_CLIENT_FACTORY format. Use 'module_path:factory_name'."
        )
    module_name, symbol = factory_spec.split(":", 1)
    module = importlib.import_module(module_name)
    target = getattr(module, symbol)
    client = target() if callable(target) else target
    if not hasattr(client, "call_tool"):
        raise RuntimeError("Loaded MCP client does not implement call_tool(server, tool, arguments).")
    return client


class PaperBananaClient:
    """
    Minimal MCP wrapper for PaperBanana. This class does not reimplement
    PaperBanana; it delegates to an MCP server tool call.
    """

    def __init__(self, mcp_client: Optional[Any] = None, server: Optional[str] = None) -> None:
        self.server = server or os.getenv("PAPERFIG_MCP_SERVER")
        self.mock_mode = os.getenv("PAPERFIG_MOCK_PAPERBANANA", "0") == "1"
        self.mcp_client = mcp_client or self._resolve_mcp_client()

    def generate_svg(self, spec: Dict[str, Any]) -> Tuple[str, List[Dict[str, Any]]]:
        if self.mock_mode:
            return self._mock_svg(spec)

        if not self.server or not self.mcp_client:
            raise RuntimeError(
                "PaperBanana MCP is not configured. Set PAPERFIG_MCP_SERVER and either "
                "PAPERFIG_MCP_CLIENT_FACTORY or PAPERFIG_MCP_COMMAND, or enable PAPERFIG_MOCK_PAPERBANANA=1."
            )

        # Expected MCP tool signature: paperbanana.generate
        response = self.mcp_client.call_tool(
            server=self.server,
            tool="paperbanana.generate",
            arguments=spec,
        )
        svg = response.get("svg")
        elements = response.get("elements", [])
        if not svg:
            raise RuntimeError("PaperBanana MCP did not return SVG output.")
        return svg, elements

    def _resolve_mcp_client(self) -> Optional[Any]:
        factory = os.getenv("PAPERFIG_MCP_CLIENT_FACTORY")
        if factory:
            return _load_client_factory(factory)

        command = os.getenv("PAPERFIG_MCP_COMMAND")
        if command:
            args = shlex.split(os.getenv("PAPERFIG_MCP_ARGS", ""))
            return PythonSDKMCPClient(command=command, args=args)

        return None

    def _mock_svg(self, spec: Dict[str, Any]) -> Tuple[str, List[Dict[str, Any]]]:
        figure_id = spec.get("figure_id") or f"fig-{uuid.uuid4().hex[:8]}"
        title = spec.get("title", "Figure")
        svg = (
            "<svg xmlns='http://www.w3.org/2000/svg' width='800' height='450' viewBox='0 0 800 450'>"
            "<rect x='40' y='40' width='720' height='370' fill='white' stroke='black'/>"
            f"<text x='60' y='90' font-family='Times New Roman' font-size='24'>{title}</text>"
            "<line x1='60' y1='120' x2='740' y2='120' stroke='black'/>"
            "<text x='60' y='170' font-family='Times New Roman' font-size='16'>Mock PaperBanana output</text>"
            "</svg>"
        )
        elements = [
            {
                "id": f"{figure_id}-title",
                "type": "text",
                "label": title,
                "source_spans": spec.get("source_spans", []),
            }
        ]
        return svg, elements



from __future__ import annotations

import importlib
import json
import os
import platform
from pathlib import Path
import shlex
import shutil
from typing import List, Optional

import typer

try:
    from paperfig import __version__
except Exception:
    try:
        from importlib.metadata import version as _pkg_version

        __version__ = _pkg_version("paperfigg")
    except Exception:
        __version__ = "0.4.0"

PACKAGE_NAME = "paperfigg"
from paperfig.agents.architecture_critic import SEVERITY_ORDER
from paperfig.agents.critic import CriticAgent
from paperfig.command_catalog import get_command_catalog
from paperfig.lab.orchestrator import LabOrchestrator
from paperfig.pipeline.orchestrator import Orchestrator
from paperfig.plugins.registry import list_plugins as list_registered_plugins
from paperfig.plugins.registry import validate_plugins as validate_registered_plugins
from paperfig.templates.lint import lint_template_catalog
from paperfig.templates.loader import load_template_catalog, validate_template_catalog
from paperfig.utils.config import load_config
from paperfig.journals import load_journal_profile
from paperfig.utils.paperbanana import PaperBananaClient
from paperfig.utils.pdf_parser import parse_paper
from paperfig.utils.types import FigurePlan, PaperContent, PaperSection

app = typer.Typer(help="Generate publication-ready figures from research papers.")
docs_app = typer.Typer(help="Documentation regeneration and drift checks.")
templates_app = typer.Typer(help="Flow template utilities.")
lab_app = typer.Typer(help="Autonomous research lab workflows.")
plugins_app = typer.Typer(help="Plugin registry utilities.")

app.add_typer(docs_app, name="docs")
app.add_typer(templates_app, name="templates")
app.add_typer(lab_app, name="lab")
app.add_typer(plugins_app, name="plugins")


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        False,
        "--version",
        callback=_version_callback,
        is_eager=True,
        help="Show paperfig version and exit.",
    ),
) -> None:
    del version


def _lab_orchestrator() -> LabOrchestrator:
    config = load_config()
    lab_cfg = config.get("lab", {})
    return LabOrchestrator(
        root_dir=Path(str(lab_cfg.get("registry_dir", "lab_runs"))),
        policy_path=Path(str(lab_cfg.get("sandbox_policy", "config/lab_policy.yaml"))),
        runs_root=Path("runs"),
    )


def _dependency_check(module_name: str, required: bool) -> dict:
    try:
        importlib.import_module(module_name)
        return {
            "check": f"python_module:{module_name}",
            "status": "ok",
            "required": required,
            "message": f"Module '{module_name}' is importable.",
        }
    except Exception as exc:
        status = "fail" if required else "warn"
        message = f"Module '{module_name}' is unavailable: {exc}"
        if module_name == "cairosvg":
            message = f"{message} Run: paperfig doctor --fix png"
        return {
            "check": f"python_module:{module_name}",
            "status": status,
            "required": required,
            "message": message,
        }


def _png_fix_guidance() -> str:
    system = platform.system().lower()
    if system.startswith("windows"):
        return "\n".join(
            [
                "Windows PNG setup options:",
                "1) MSYS2 UCRT64 (recommended for system-wide Cairo):",
                "   - winget install MSYS2.MSYS2",
                "   - C:\\msys64\\ucrt64.exe",
                "   - pacman -Syu",
                "   - pacman -S --needed mingw-w64-ucrt-x86_64-cairo mingw-w64-ucrt-x86_64-pango mingw-w64-ucrt-x86_64-gdk-pixbuf2",
                "   - Add C:\\msys64\\ucrt64\\bin to PATH",
                "2) Conda-forge environment:",
                "   - conda create -n paperfig-png python=3.10 cairosvg cairo pango gdk-pixbuf -c conda-forge",
                "   - conda activate paperfig-png",
                "Verification:",
                "   - paperfig doctor --fix png --verify",
                "   - paperfig doctor",
            ]
        )
    return "\n".join(
        [
            "Install Cairo system libraries and cairosvg for PNG export.",
            "Then verify with:",
            "  - paperfig doctor --fix png --verify",
            "  - paperfig doctor",
        ]
    )


def _mcp_check(probe_mcp: bool) -> dict:
    if os.getenv("PAPERFIG_MOCK_PAPERBANANA", "0") == "1":
        return {
            "check": "paperbanana_mcp",
            "status": "ok",
            "required": False,
            "message": "Mock mode is enabled (PAPERFIG_MOCK_PAPERBANANA=1); real MCP probe skipped.",
        }

    server = os.getenv("PAPERFIG_MCP_SERVER", "").strip()
    factory = os.getenv("PAPERFIG_MCP_CLIENT_FACTORY", "").strip()
    command = os.getenv("PAPERFIG_MCP_COMMAND", "").strip()

    if not server:
        return {
            "check": "paperbanana_mcp",
            "status": "warn",
            "required": False,
            "message": "PAPERFIG_MCP_SERVER is not configured.",
        }

    if not factory and not command:
        return {
            "check": "paperbanana_mcp",
            "status": "warn",
            "required": False,
            "message": "No MCP transport configured. Set PAPERFIG_MCP_CLIENT_FACTORY or PAPERFIG_MCP_COMMAND.",
        }

    if command:
        command_parts = shlex.split(command)
        command_bin = command_parts[0] if command_parts else ""
        if not command_bin:
            return {
                "check": "paperbanana_mcp",
                "status": "fail",
                "required": False,
                "message": "PAPERFIG_MCP_COMMAND is set but empty after parsing.",
            }
        if shutil.which(command_bin) is None:
            return {
                "check": "paperbanana_mcp",
                "status": "fail",
                "required": False,
                "message": f"MCP command binary '{command_bin}' is not on PATH.",
            }

    if not probe_mcp:
        return {
            "check": "paperbanana_mcp",
            "status": "ok",
            "required": False,
            "message": "MCP configuration detected. Run `paperfig doctor --probe-mcp` for an end-to-end probe.",
        }

    try:
        client = PaperBananaClient()
        client.generate_svg(
            {
                "figure_id": "doctor-probe",
                "title": "doctor-probe",
                "kind": "system_overview",
                "description": "Connectivity probe",
                "abstraction_level": "high",
                "source_text": {"methodology": "", "system": "", "results": ""},
                "source_spans": [],
                "style_refs": {},
                "critique_feedback": {},
                "iteration": 1,
            }
        )
        return {
            "check": "paperbanana_mcp",
            "status": "ok",
            "required": False,
            "message": "MCP probe succeeded via `paperbanana.generate`.",
        }
    except Exception as exc:
        return {
            "check": "paperbanana_mcp",
            "status": "fail",
            "required": False,
            "message": f"MCP probe failed: {exc}",
        }


def _render_doctor_output(report: dict) -> None:
    checks = report.get("checks", [])
    try:
        from rich.console import Console
        from rich.table import Table

        table = Table(title="paperfig doctor")
        table.add_column("Check", justify="left")
        table.add_column("Status", justify="left")
        table.add_column("Required", justify="center")
        table.add_column("Message", justify="left")

        status_style = {"ok": "green", "warn": "yellow", "fail": "red"}
        for check in checks:
            status = str(check.get("status", "warn"))
            table.add_row(
                str(check.get("check", "")),
                f"[{status_style.get(status, 'white')}]{status}[/{status_style.get(status, 'white')}]",
                "yes" if check.get("required", False) else "no",
                str(check.get("message", "")),
            )

        console = Console()
        console.print(table)
        console.print(f"Package: {PACKAGE_NAME} | Command: paperfig")
        console.print(
            f"Overall: {'PASS' if report.get('passed') else 'FAIL'} "
            f"(required failures: {report.get('required_failures', 0)}, "
            f"optional failures: {report.get('optional_failures', 0)})"
        )
        if any("paperfig doctor --fix png" in str(check.get("message", "")) for check in checks):
            console.print("Hint: paperfig doctor --fix png")
    except Exception:
        typer.echo("paperfig doctor")
        for check in checks:
            typer.echo(
                f"- {check.get('check')}: {check.get('status')} "
                f"(required={check.get('required')}) {check.get('message')}"
            )
        typer.echo(f"Package: {PACKAGE_NAME} | Command: paperfig")
        typer.echo(
            "Overall: "
            f"{'PASS' if report.get('passed') else 'FAIL'} "
            f"(required failures: {report.get('required_failures', 0)}, "
            f"optional failures: {report.get('optional_failures', 0)})"
        )
        if any("paperfig doctor --fix png" in str(check.get("message", "")) for check in checks):
            typer.echo("Hint: paperfig doctor --fix png")


@app.command()
def generate(
    paper_path: Path = typer.Argument(..., exists=True, help="Path to paper PDF or Markdown"),
    mode: str = typer.Option(
        "auto",
        "--mode",
        help="PaperBanana execution mode: auto, mock, real, or journal:<profile>.",
    ),
    run_root: Path = typer.Option(Path("runs"), help="Root directory for run outputs"),
    max_iterations: int = typer.Option(3, help="Maximum generation iterations per figure"),
    quality_threshold: float = typer.Option(0.75, help="Minimum critique score to accept a figure"),
    dimension_threshold: float = typer.Option(
        0.55,
        help="Minimum score required for each critique dimension.",
    ),
    template_pack: Optional[str] = typer.Option(
        None,
        help="Template pack ID for planner flow templates.",
    ),
    arch_critique_mode: str = typer.Option(
        "inline",
        help="Architecture critique mode: inline or off.",
    ),
    arch_critique_block_severity: str = typer.Option(
        "critical",
        help="Minimum architecture finding severity that blocks generation.",
    ),
    repro_audit_mode: str = typer.Option(
        "soft",
        help="Reproducibility audit mode: soft or hard.",
    ),
    contrib: bool = typer.Option(
        False,
        "--contrib",
        help="Contributor mode: verbose artifacts, planner/critic notes, and run CONTRIBUTING_NOTES.",
    ),
) -> None:
    journal_profile = None
    if mode.startswith("journal:"):
        profile_id = mode.split(":", 1)[1].strip()
        if not profile_id:
            raise typer.BadParameter("journal mode requires a profile name, e.g. journal:neurips")
        journal_profile = load_journal_profile(profile_id)
        max_iterations = journal_profile.max_iterations
        quality_threshold = journal_profile.quality_threshold
        dimension_threshold = journal_profile.dimension_threshold
        template_pack = journal_profile.template_pack or template_pack
        arch_critique_block_severity = journal_profile.arch_critique_block_severity
        repro_audit_mode = journal_profile.repro_audit_mode
        mode = "auto"
    elif mode not in {"auto", "mock", "real"}:
        raise typer.BadParameter("mode must be one of: auto, mock, real, journal:<profile>")
    if arch_critique_mode not in {"inline", "off"}:
        raise typer.BadParameter("arch_critique_mode must be one of: inline, off")
    if arch_critique_block_severity not in SEVERITY_ORDER:
        raise typer.BadParameter(
            f"arch_critique_block_severity must be one of: {', '.join(SEVERITY_ORDER.keys())}"
        )
    if repro_audit_mode not in {"soft", "hard"}:
        raise typer.BadParameter("repro_audit_mode must be one of: soft, hard")

    previous_mock_mode = os.getenv("PAPERFIG_MOCK_PAPERBANANA")
    if mode == "mock":
        os.environ["PAPERFIG_MOCK_PAPERBANANA"] = "1"
    elif mode == "real":
        os.environ["PAPERFIG_MOCK_PAPERBANANA"] = "0"

    try:
        orchestrator = Orchestrator(
            run_root=run_root,
            max_iterations=max_iterations,
            quality_threshold=quality_threshold,
            dimension_threshold=dimension_threshold,
            template_pack=template_pack,
            arch_critique_mode=arch_critique_mode,
            arch_critique_block_severity=arch_critique_block_severity,
            repro_audit_mode=repro_audit_mode,
            journal_profile=journal_profile,
        )
        run_id = orchestrator.generate(paper_path, contrib=contrib)
    except RuntimeError as exc:
        typer.echo(f"Generation failed: {exc}")
        raise typer.Exit(code=1)
    finally:
        if mode in {"mock", "real"}:
            if previous_mock_mode is None:
                os.environ.pop("PAPERFIG_MOCK_PAPERBANANA", None)
            else:
                os.environ["PAPERFIG_MOCK_PAPERBANANA"] = previous_mock_mode

    typer.echo(f"Run created: {run_id}")
    typer.echo(f"Output directory: {run_root / run_id}")


@app.command()
def critique(
    figure_path: Path = typer.Argument(..., exists=True, help="Path to figure SVG"),
    paper_path: Optional[Path] = typer.Option(None, help="Optional paper PDF or Markdown for context"),
    threshold: float = typer.Option(0.75, help="Critique acceptance threshold"),
    dimension_threshold: float = typer.Option(
        0.55,
        help="Minimum score required for each critique dimension.",
    ),
) -> None:
    if paper_path:
        paper = parse_paper(paper_path)
    else:
        paper = PaperContent(
            source_path="",
            full_text="",
            sections={
                "methodology": PaperSection("methodology", "", 0, 0),
                "system": PaperSection("system", "", 0, 0),
                "results": PaperSection("results", "", 0, 0),
            },
        )

    plan = FigurePlan(
        figure_id=figure_path.stem,
        title=figure_path.stem,
        kind="ad_hoc",
        order=1,
        abstraction_level="unknown",
        description="Ad-hoc critique",
        justification="Manual critique invocation",
        template_id="manual",
        source_spans=[],
    )

    critic = CriticAgent(threshold=threshold, dimension_threshold=dimension_threshold)
    report = critic.critique(figure_path, plan, paper)
    typer.echo(json.dumps(report.__dict__, indent=2))


@app.command()
def rerun(
    run_id: str = typer.Argument(..., help="Existing run ID to replay deterministically"),
    run_root: Path = typer.Option(Path("runs"), help="Root directory for run outputs"),
    contrib: bool = typer.Option(
        False,
        "--contrib",
        help="Contributor mode for the replayed run.",
    ),
) -> None:
    orchestrator = Orchestrator(run_root=run_root)
    try:
        new_run_id = orchestrator.rerun(source_run_id=run_id, contrib=contrib)
    except RuntimeError as exc:
        typer.echo(f"Rerun failed: {exc}")
        raise typer.Exit(code=1)
    typer.echo(f"Rerun created: {new_run_id}")
    typer.echo(f"Output directory: {run_root / new_run_id}")


@app.command()
def diff(
    run_id_1: str = typer.Argument(..., help="Base run ID"),
    run_id_2: str = typer.Argument(..., help="Target run ID"),
    run_root: Path = typer.Option(Path("runs"), help="Root directory for run outputs"),
    output_dir: Optional[Path] = typer.Option(None, help="Optional output directory for diff artifacts"),
    as_json: bool = typer.Option(False, "--as-json", help="Print diff report as JSON"),
) -> None:
    orchestrator = Orchestrator(run_root=run_root)
    report = orchestrator.diff(run_id_1=run_id_1, run_id_2=run_id_2, output_dir=output_dir)
    if as_json:
        typer.echo(json.dumps(report, indent=2))
        return

    typer.echo(f"Run 1: {run_id_1}")
    typer.echo(f"Run 2: {run_id_2}")
    typer.echo(f"Diff output: {Path(report.get('diff_dir', '')) / 'diff.json'}")
    metrics = report.get("metrics", {})
    typer.echo(
        "accepted_count: "
        f"{metrics.get('accepted_count', {}).get('run_1')} -> "
        f"{metrics.get('accepted_count', {}).get('run_2')}"
    )
    typer.echo(
        "avg_final_score: "
        f"{metrics.get('avg_final_score', {}).get('run_1')} -> "
        f"{metrics.get('avg_final_score', {}).get('run_2')}"
    )
    typer.echo(
        "avg_traceability_coverage: "
        f"{metrics.get('avg_traceability_coverage', {}).get('run_1')} -> "
        f"{metrics.get('avg_traceability_coverage', {}).get('run_2')}"
    )
    typer.echo(f"Changed figures: {len(report.get('changed_figures', []))}")
    typer.echo(f"Changed JSON artifacts: {len(report.get('changed_artifacts', []))}")


@app.command()
def regress(
    paper_v1: Path = typer.Argument(..., exists=True, help="Baseline paper (v1) path"),
    paper_v2: Path = typer.Argument(..., exists=True, help="Comparison paper (v2) path"),
    mode: str = typer.Option("mock", "--mode", help="PaperBanana execution mode: auto, mock, or real."),
    run_root: Path = typer.Option(Path("runs"), help="Root directory for run outputs"),
    output_dir: Optional[Path] = typer.Option(None, help="Optional output directory for regression report"),
    as_json: bool = typer.Option(False, "--as-json", help="Print regression report as JSON"),
) -> None:
    if mode not in {"auto", "mock", "real"}:
        raise typer.BadParameter("mode must be one of: auto, mock, real")

    previous_mock_mode = os.getenv("PAPERFIG_MOCK_PAPERBANANA")
    if mode == "mock":
        os.environ["PAPERFIG_MOCK_PAPERBANANA"] = "1"
    elif mode == "real":
        os.environ["PAPERFIG_MOCK_PAPERBANANA"] = "0"

    try:
        orchestrator = Orchestrator(run_root=run_root)
        report = orchestrator.regress(paper_v1, paper_v2, output_dir=output_dir)
    finally:
        if mode in {"mock", "real"}:
            if previous_mock_mode is None:
                os.environ.pop("PAPERFIG_MOCK_PAPERBANANA", None)
            else:
                os.environ["PAPERFIG_MOCK_PAPERBANANA"] = previous_mock_mode

    if as_json:
        typer.echo(json.dumps(report, indent=2))
        return

    typer.echo(f"Regression report: {Path(report.get('report_dir', '')) / 'regression_report.json'}")
    typer.echo(report.get("summary", ""))


@app.command()
def export(
    run_id: str = typer.Argument(..., help="Run ID to export"),
    run_root: Path = typer.Option(Path("runs"), help="Root directory for run outputs"),
    output_dir: Optional[Path] = typer.Option(None, help="Optional export output directory"),
) -> None:
    orchestrator = Orchestrator(run_root=run_root)
    export_root = orchestrator.export(run_id, output_dir=output_dir)
    typer.echo(f"Exports written to: {export_root}")
    report_path = export_root / "export_report.json"
    if report_path.exists():
        report = json.loads(report_path.read_text(encoding="utf-8"))
        for warning in report.get("warnings", []):
            typer.echo(f"Warning: {warning}")


@app.command()
def doctor(
    as_json: bool = typer.Option(False, "--as-json", help="Print doctor report as JSON."),
    probe_mcp: bool = typer.Option(
        False,
        "--probe-mcp",
        help="Attempt an end-to-end PaperBanana MCP probe.",
    ),
    strict: bool = typer.Option(
        False,
        "--strict",
        help="Exit non-zero if any check reports failure.",
    ),
    fix: Optional[str] = typer.Option(
        None,
        "--fix",
        help="Show guided fix instructions for a subsystem (currently: png).",
    ),
    verify: bool = typer.Option(
        False,
        "--verify",
        help="Re-run targeted verification for the selected --fix subsystem.",
    ),
) -> None:
    if fix and fix not in {"png"}:
        raise typer.BadParameter("fix must be one of: png")

    checks = [
        _dependency_check("typer", required=True),
        _dependency_check("rich", required=False),
        _dependency_check("cairosvg", required=False),
        _mcp_check(probe_mcp=probe_mcp),
    ]

    required_failures = sum(
        1 for check in checks if check.get("required", False) and check.get("status") == "fail"
    )
    optional_failures = sum(
        1 for check in checks if not check.get("required", False) and check.get("status") == "fail"
    )

    report = {
        "version": __version__,
        "checks": checks,
        "probe_mcp": probe_mcp,
        "required_failures": required_failures,
        "optional_failures": optional_failures,
        "passed": required_failures == 0,
    }

    if as_json:
        typer.echo(json.dumps(report, indent=2))
    else:
        _render_doctor_output(report)
        if fix == "png":
            typer.echo("")
            typer.echo(_png_fix_guidance())
            if verify:
                verify_check = _dependency_check("cairosvg", required=False)
                typer.echo("")
                typer.echo(
                    "PNG verify: "
                    f"{verify_check.get('status')} - {verify_check.get('message')}"
                )

    if required_failures > 0:
        raise typer.Exit(code=1)
    if strict and optional_failures > 0:
        raise typer.Exit(code=1)


@app.command(name="critique-architecture")
def critique_architecture(
    run_id: Optional[str] = typer.Argument(None, help="Run ID to critique"),
    run_root: Path = typer.Option(Path("runs"), help="Root directory for run outputs"),
    as_json: bool = typer.Option(False, "--as-json", help="Print report as JSON"),
    block_severity: str = typer.Option(
        "critical",
        help="Severity threshold that marks the report as blocking.",
    ),
    list_rules: bool = typer.Option(False, "--list-rules", help="List built-in architecture rules and exit."),
    enable: Optional[List[str]] = typer.Option(
        None,
        "--enable",
        help="Enable a subset of architecture rules (repeatable). Default: all built-in rules.",
    ),
) -> None:
    if block_severity not in SEVERITY_ORDER:
        raise typer.BadParameter(
            f"block_severity must be one of: {', '.join(SEVERITY_ORDER.keys())}"
        )

    orchestrator = Orchestrator(run_root=run_root)
    if list_rules:
        for rule in orchestrator.architecture_critic.available_rules():
            typer.echo(f"- {rule['rule_id']}: {rule['description']}")
        return
    if not run_id:
        raise typer.BadParameter("run_id is required unless --list-rules is provided")

    try:
        report = orchestrator.critique_architecture(
            run_id=run_id,
            block_severity=block_severity,
            persist=True,
            enabled_rules=enable,
        )
    except ValueError as exc:
        raise typer.BadParameter(str(exc)) from exc

    if as_json:
        typer.echo(json.dumps(report, indent=2))
        return

    typer.echo(f"Run: {run_id}")
    typer.echo(f"Summary: {report.get('summary')}")
    typer.echo(f"Blocked: {report.get('blocked')}")
    findings = report.get("findings", [])
    for finding in findings:
        typer.echo(f"- [{finding.get('severity')}] {finding.get('title')}: {finding.get('description')}")


@app.command()
def audit(
    run_id: str = typer.Argument(..., help="Run ID to audit"),
    run_root: Path = typer.Option(Path("runs"), help="Root directory for run outputs"),
    mode: str = typer.Option("soft", help="Audit mode: soft or hard"),
    as_json: bool = typer.Option(False, "--as-json", help="Print report as JSON"),
) -> None:
    if mode not in {"soft", "hard"}:
        raise typer.BadParameter("mode must be one of: soft, hard")

    orchestrator = Orchestrator(run_root=run_root)
    report = orchestrator.audit(run_id=run_id, mode=mode, persist=True)

    if as_json:
        typer.echo(json.dumps(report, indent=2))
    else:
        typer.echo(f"Run: {run_id}")
        typer.echo(f"Mode: {mode}")
        typer.echo(f"Passed: {report.get('passed')}")
        typer.echo(f"Summary: {report.get('summary')}")

    if mode == "hard" and not report.get("passed", False):
        raise typer.Exit(code=1)


@app.command()
def inspect(
    run_id: str = typer.Argument(..., help="Run ID to inspect"),
    run_root: Path = typer.Option(Path("runs"), help="Root directory for run outputs"),
    as_json: bool = typer.Option(False, "--as-json", help="Print the full summary JSON"),
    html: bool = typer.Option(False, "--html", help="Generate a self-contained HTML inspector."),
    failures_only: bool = typer.Option(
        False,
        "--failures-only",
        help="Only include figures that did not pass final critique.",
    ),
    figure_id: Optional[str] = typer.Option(
        None,
        "--figure-id",
        help="Only include a specific figure ID.",
    ),
    min_score: Optional[float] = typer.Option(
        None,
        "--min-score",
        help="Only include figures with final score >= this value.",
    ),
    failed_dimension: Optional[str] = typer.Option(
        None,
        "--failed-dimension",
        help="Only include figures that failed a specific dimension.",
    ),
    output_path: Optional[Path] = typer.Option(None, help="Optional path to write summary JSON"),
) -> None:
    orchestrator = Orchestrator(run_root=run_root)
    summary = orchestrator.inspect(
        run_id,
        failures_only=failures_only,
        figure_id=figure_id,
        min_score=min_score,
        failed_dimension=failed_dimension,
    )

    if output_path:
        output_path.parent.mkdir(parents=True, exist_ok=True)
        output_path.write_text(json.dumps(summary, indent=2), encoding="utf-8")
        typer.echo(f"Summary written to: {output_path}")

    if html:
        html_path = orchestrator.inspect_html(run_id)
        typer.echo(f"HTML inspector written to: {html_path}")

    if as_json:
        typer.echo(json.dumps(summary, indent=2))
        return

    aggregate = summary.get("aggregate", {})
    metadata = summary.get("metadata", {})
    typer.echo(f"Run: {summary.get('run_id')}")
    if metadata.get("paper_path"):
        typer.echo(f"Paper: {metadata['paper_path']}")
    typer.echo(
        "Figures: "
        f"{aggregate.get('accepted_count', 0)}/{aggregate.get('total_figures', 0)} accepted, "
        f"{aggregate.get('failed_count', 0)} failed"
    )
    typer.echo(
        "Averages: "
        f"score={aggregate.get('avg_final_score')} "
        f"traceability_coverage={aggregate.get('avg_traceability_coverage')}"
    )

    for figure in summary.get("figures", []):
        coverage = figure.get("traceability", {}).get("coverage")
        coverage_text = f"{coverage:.2f}" if isinstance(coverage, (int, float)) else "n/a"
        typer.echo(
            f"- {figure.get('figure_id')} "
            f"passed={figure.get('final_passed')} "
            f"iter={figure.get('iterations_attempted')} "
            f"score={figure.get('final_score')} "
            f"coverage={coverage_text}"
        )
        issues = figure.get("issues") or []
        if issues:
            typer.echo(f"  issue: {issues[0]}")

    for warning in summary.get("warnings", []):
        typer.echo(f"Warning: {warning}")


@docs_app.command("regenerate")
def docs_regenerate(
    check: bool = typer.Option(False, "--check", help="Check for drift without applying changes."),
    report_path: Optional[Path] = typer.Option(None, help="Optional path to write drift report JSON."),
) -> None:
    orchestrator = Orchestrator()
    report = orchestrator.docs_regenerate(check_only=check, report_path=report_path)

    typer.echo(f"Checked documents: {len(report.get('documents', []))}")
    typer.echo(f"Drift detected: {report.get('drift_detected')}")

    if check and report.get("drift_detected"):
        raise typer.Exit(code=1)


@docs_app.command("check")
def docs_check(
    report_path: Optional[Path] = typer.Option(None, help="Optional path to write drift report JSON."),
) -> None:
    docs_regenerate(check=True, report_path=report_path)


@templates_app.command("list")
def templates_list(
    template_dir: Path = typer.Option(Path("paperfig/templates/flows"), help="Template directory."),
    pack_id: str = typer.Option("expanded_v1", help="Template pack identifier."),
    pack: Optional[str] = typer.Option(
        None,
        "--pack",
        help="External template pack source (directory path or python package).",
    ),
) -> None:
    catalog = load_template_catalog(template_dir=template_dir, pack_id=pack_id, pack=pack)
    typer.echo(f"Template pack: {catalog.pack_id}")
    for template in catalog.templates:
        typer.echo(f"- {template.template_id}: {template.title} ({template.kind})")


@templates_app.command("validate")
def templates_validate(
    template_dir: Path = typer.Option(Path("paperfig/templates/flows"), help="Template directory."),
    pack_id: str = typer.Option("expanded_v1", help="Template pack identifier."),
    pack: Optional[str] = typer.Option(
        None,
        "--pack",
        help="External template pack source (directory path or python package).",
    ),
) -> None:
    errors = validate_template_catalog(template_dir=template_dir, pack_id=pack_id, pack=pack)
    if errors:
        for error in errors:
            typer.echo(f"Error: {error}")
        raise typer.Exit(code=1)
    typer.echo("Template catalog is valid.")


@templates_app.command("lint")
def templates_lint(
    template_dir: Path = typer.Option(Path("paperfig/templates/flows"), help="Template directory."),
    pack: Optional[str] = typer.Option(
        None,
        "--pack",
        help="External template pack source (directory path or python package).",
    ),
) -> None:
    errors = lint_template_catalog(template_dir=template_dir, pack=pack)
    if errors:
        for error in errors:
            typer.echo(f"Error: {error}")
        raise typer.Exit(code=1)
    typer.echo("Flow templates satisfy flow_template.schema.json.")


@plugins_app.command("list")
def plugins_list(
    kind: Optional[str] = typer.Option(
        None,
        help="Optional plugin kind filter: critique_rule or repro_check.",
    ),
) -> None:
    plugins = list_registered_plugins(kind=kind)
    if not plugins:
        typer.echo("No plugins found.")
        return
    for plugin in plugins:
        typer.echo(f"- {plugin.plugin_id} ({plugin.kind}): {plugin.description}")


@plugins_app.command("validate")
def plugins_validate(
    kind: Optional[str] = typer.Option(
        None,
        help="Optional plugin kind filter: critique_rule or repro_check.",
    ),
) -> None:
    errors = validate_registered_plugins(kind=kind)
    if errors:
        for error in errors:
            typer.echo(f"Error: {error}")
        raise typer.Exit(code=1)
    typer.echo("Plugin registry is valid.")


@lab_app.command("init")
def lab_init() -> None:
    orchestrator = _lab_orchestrator()
    lab_run_id = orchestrator.init_lab()
    typer.echo(f"Initialized lab run: {lab_run_id}")


@lab_app.command("propose")
def lab_propose(topic_or_run_id: str = typer.Argument(..., help="Topic text or existing run ID")) -> None:
    orchestrator = _lab_orchestrator()
    spec = orchestrator.propose(topic_or_run_id)
    typer.echo(f"Proposed experiment: {spec.experiment_id}")
    typer.echo(f"Command: {spec.command}")


@lab_app.command("run")
def lab_run(experiment_id: str = typer.Argument(..., help="Experiment ID")) -> None:
    orchestrator = _lab_orchestrator()
    result = orchestrator.run(experiment_id)
    typer.echo(f"Experiment: {experiment_id}")
    typer.echo(f"Status: {result.status}")
    typer.echo(f"Return code: {result.return_code}")
    if result.policy_violation:
        typer.echo(f"Policy violation: {result.policy_violation}")


@lab_app.command("review")
def lab_review(experiment_id: str = typer.Argument(..., help="Experiment ID")) -> None:
    orchestrator = _lab_orchestrator()
    review = orchestrator.review(experiment_id)
    typer.echo(f"Experiment: {experiment_id}")
    typer.echo(f"Recommendation: {review.get('recommendation')}")
    typer.echo(f"Rationale: {review.get('rationale')}")


@lab_app.command("status")
def lab_status(
    run_id: Optional[str] = typer.Option(None, "--run-id", help="Optional lab run ID"),
) -> None:
    orchestrator = _lab_orchestrator()
    status = orchestrator.status(lab_run_id=run_id)
    typer.echo(f"Lab run: {status.get('lab_run_id')}")
    typer.echo(json.dumps(status.get("counts", {}), indent=2))
    for item in status.get("experiments", []):
        typer.echo(f"- {item.get('experiment_id')} status={item.get('status')} topic={item.get('topic')}")


@app.command("command-catalog")
def command_catalog() -> None:
    for command in get_command_catalog():
        typer.echo(command)


if __name__ == "__main__":
    app()



from __future__ import annotations

from typing import List

from paperfig.critique.rules.base import RuleContext
from paperfig.utils.types import ArchitectureCritiqueFinding


RULE_ID = "empty_plan"
DESCRIPTION = "Reject runs where planner emitted no figure plans."


def evaluate(context: RuleContext) -> List[ArchitectureCritiqueFinding]:
    if context.plan_data is None:
        return []
    if context.plan_data:
        return []
    return [
        ArchitectureCritiqueFinding(
            finding_id=RULE_ID,
            severity="critical",
            title="Empty figure plan",
            description="Planner generated no figures for this run.",
            evidence="plan.json contains 0 entries",
            suggestion="Review section extraction and template trigger rules.",
        )
    ]




from __future__ import annotations


def propose_hypothesis(topic: str, source_run_id: str = "") -> str:
    if source_run_id:
        return (
            f"Adjust figure planning and critique thresholds for run {source_run_id} "
            "to improve traceability coverage and acceptance rate."
        )
    return f"Investigate whether template-guided generation improves output quality for topic: {topic}."



from __future__ import annotations

import tempfile
import unittest
from pathlib import Path

from paperfig.utils.pdf_parser import parse_paper


class ParserTests(unittest.TestCase):
    def test_parse_markdown_extracts_required_sections(self) -> None:
        content = """
# Title

## 1 Methodology
We optimize with a staged pipeline.

## 2 System Architecture
The architecture uses parser, planner, generator, critic.

## 3 Results
We improve accuracy by 7 points.
""".strip()

        with tempfile.TemporaryDirectory() as tmpdir:
            path = Path(tmpdir) / "paper.md"
            path.write_text(content, encoding="utf-8")
            parsed = parse_paper(path)

        self.assertIn("methodology", parsed.sections)
        self.assertIn("system", parsed.sections)
        self.assertIn("results", parsed.sections)
        self.assertGreater(len(parsed.sections["methodology"].text), 0)
        self.assertGreater(len(parsed.sections["system"].text), 0)
        self.assertGreater(len(parsed.sections["results"].text), 0)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, Optional

from paperfig.utils.types import ReproAuditCheck


@dataclass(frozen=True)
class ReproCheckDefinition:
    check_id: str
    description: str
    required: bool
    severity: str
    evaluator: Callable[[Path, Optional[str]], ReproAuditCheck]


def _artifact_check(run_dir: Path, relative_path: str, required: bool = True) -> ReproAuditCheck:
    path = run_dir / relative_path
    exists = path.exists()
    return ReproAuditCheck(
        check_id=f"artifact_{relative_path.replace('/', '_')}",
        description=f"Artifact exists: {relative_path}",
        required=required,
        passed=exists,
        severity="major" if required else "minor",
        message="present" if exists else "missing",
        details={"path": str(path)},
    )


def _load_run_json(run_dir: Path) -> tuple[Dict[str, object], bool]:
    run_json_path = run_dir / "run.json"
    if not run_json_path.exists():
        return {}, False
    return json.loads(run_json_path.read_text(encoding="utf-8")), True


def _check_run_json_present(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    run_json_path = run_dir / "run.json"
    exists = run_json_path.exists()
    return ReproAuditCheck(
        check_id="run_json_present",
        description="Run metadata file exists",
        required=True,
        passed=exists,
        severity="critical",
        message="present" if exists else "missing",
        details={"path": str(run_json_path)},
    )


def _check_plan(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    return _artifact_check(run_dir, "plan.json", required=True)


def _check_sections(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    return _artifact_check(run_dir, "sections.json", required=True)


def _check_traceability(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    return _artifact_check(run_dir, "traceability.json", required=True)


def _check_inspect(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    return _artifact_check(run_dir, "inspect.json", required=True)


def _check_docs_drift(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    return _artifact_check(run_dir, "docs_drift_report.json", required=True)


def _check_architecture_critique(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    return _artifact_check(run_dir, "architecture_critique.json", required=True)


def _check_prompt_plan(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    return _artifact_check(run_dir, "prompts/plan_figure.txt", required=True)


def _check_prompt_critique(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    return _artifact_check(run_dir, "prompts/critique_figure.txt", required=True)


def _check_provenance(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    run_json, has_run_json = _load_run_json(run_dir)
    has_command_meta = bool(run_json.get("paper_path")) and bool(run_json.get("created_at"))
    return ReproAuditCheck(
        check_id="provenance_metadata",
        description="Run metadata captures provenance fields",
        required=True,
        passed=has_run_json and has_command_meta,
        severity="major",
        message="ok" if has_command_meta else "missing_fields",
        details={"required_fields": ["paper_path", "created_at"]},
    )


def _check_seed_declared(run_dir: Path, _: Optional[str]) -> ReproAuditCheck:
    run_json, _ = _load_run_json(run_dir)
    has_seed = "seed" in run_json
    return ReproAuditCheck(
        check_id="deterministic_seed_declared",
        description="Run metadata declares a deterministic seed",
        required=False,
        passed=has_seed,
        severity="minor",
        message="seed_present" if has_seed else "seed_missing",
        details={},
    )


def _check_config_hash(run_dir: Path, expected_config_hash: Optional[str]) -> ReproAuditCheck:
    run_json, _ = _load_run_json(run_dir)
    if expected_config_hash is None:
        return ReproAuditCheck(
            check_id="config_hash_match",
            description="Run metadata config hash matches expected hash",
            required=True,
            passed=True,
            severity="major",
            message="skipped",
            details={"expected": None, "actual": str(run_json.get("config_hash", ""))},
        )

    run_hash = str(run_json.get("config_hash", ""))
    return ReproAuditCheck(
        check_id="config_hash_match",
        description="Run metadata config hash matches expected hash",
        required=True,
        passed=(run_hash == expected_config_hash),
        severity="major",
        message="match" if run_hash == expected_config_hash else "mismatch",
        details={"expected": expected_config_hash, "actual": run_hash},
    )


def get_repro_check_registry() -> Dict[str, ReproCheckDefinition]:
    checks = [
        ReproCheckDefinition(
            check_id="run_json_present",
            description="Run metadata file exists",
            required=True,
            severity="critical",
            evaluator=_check_run_json_present,
        ),
        ReproCheckDefinition(
            check_id="plan_present",
            description="Plan artifact exists",
            required=True,
            severity="major",
            evaluator=_check_plan,
        ),
        ReproCheckDefinition(
            check_id="sections_present",
            description="Sections artifact exists",
            required=True,
            severity="major",
            evaluator=_check_sections,
        ),
        ReproCheckDefinition(
            check_id="traceability_present",
            description="Traceability artifact exists",
            required=True,
            severity="major",
            evaluator=_check_traceability,
        ),
        ReproCheckDefinition(
            check_id="inspect_present",
            description="Inspect snapshot exists",
            required=True,
            severity="major",
            evaluator=_check_inspect,
        ),
        ReproCheckDefinition(
            check_id="docs_drift_present",
            description="Docs drift report exists",
            required=True,
            severity="major",
            evaluator=_check_docs_drift,
        ),
        ReproCheckDefinition(
            check_id="architecture_critique_present",
            description="Architecture critique report exists",
            required=True,
            severity="major",
            evaluator=_check_architecture_critique,
        ),
        ReproCheckDefinition(
            check_id="prompt_plan_present",
            description="Plan prompt exists",
            required=True,
            severity="major",
            evaluator=_check_prompt_plan,
        ),
        ReproCheckDefinition(
            check_id="prompt_critique_present",
            description="Critique prompt exists",
            required=True,
            severity="major",
            evaluator=_check_prompt_critique,
        ),
        ReproCheckDefinition(
            check_id="provenance_metadata",
            description="Run metadata captures provenance fields",
            required=True,
            severity="major",
            evaluator=_check_provenance,
        ),
        ReproCheckDefinition(
            check_id="deterministic_seed_declared",
            description="Run metadata declares a deterministic seed",
            required=False,
            severity="minor",
            evaluator=_check_seed_declared,
        ),
        ReproCheckDefinition(
            check_id="config_hash_match",
            description="Run metadata config hash matches expected hash",
            required=True,
            severity="major",
            evaluator=_check_config_hash,
        ),
    ]
    return {check.check_id: check for check in checks}



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Dict, List, Optional

from paperfig.templates.loader import discover_template_files, _normalize_template_payload
from paperfig.utils.structured_data import load_structured_file


SCHEMA_PATH = Path("paperfig/templates/schema/flow_template.schema.json")


def load_flow_template_schema(schema_path: Path = SCHEMA_PATH) -> Dict[str, Any]:
    if not schema_path.exists():
        raise FileNotFoundError(f"Flow template schema not found: {schema_path}")
    return json.loads(schema_path.read_text(encoding="utf-8"))


def _simple_schema_validate(data: Dict[str, Any], schema: Dict[str, Any]) -> List[str]:
    errors: List[str] = []
    for field in schema.get("required", []):
        if field not in data:
            errors.append(f"{field}: is required")
    return errors


def _jsonschema_validate(data: Dict[str, Any], schema: Dict[str, Any]) -> List[str]:
    try:
        from jsonschema import Draft202012Validator
    except Exception:
        return _simple_schema_validate(data, schema)

    validator = Draft202012Validator(schema)
    errors: List[str] = []
    for error in sorted(validator.iter_errors(data), key=lambda item: list(item.path)):
        field = ".".join(str(part) for part in error.path) or "<root>"
        errors.append(f"{field}: {error.message}")
    return errors


def lint_template_catalog(
    template_dir: Path,
    pack: Optional[str] = None,
    schema_path: Path = SCHEMA_PATH,
) -> List[str]:
    schema = load_flow_template_schema(schema_path=schema_path)
    paths, _, _ = discover_template_files(template_dir=template_dir, pack=pack)
    all_errors: List[str] = []
    if not paths:
        source = pack or str(template_dir)
        return [f"{source}: <root>: no template files were found"]

    for path in paths:
        data = load_structured_file(path)
        if not isinstance(data, dict):
            all_errors.append(f"{path}: <root>: template must be a mapping/object")
            continue

        normalized = _normalize_template_payload(data)
        errors = _jsonschema_validate(normalized, schema)
        for error in errors:
            all_errors.append(f"{path}: {error}")

    return all_errors



from __future__ import annotations

from typing import List

from paperfig.critique.rules.base import RuleContext
from paperfig.utils.types import ArchitectureCritiqueFinding


RULE_ID = "missing_inspect"
DESCRIPTION = "Require inspect.json so architecture quality can be assessed."


def evaluate(context: RuleContext) -> List[ArchitectureCritiqueFinding]:
    if context.inspect_data is not None:
        return []
    return [
        ArchitectureCritiqueFinding(
            finding_id=RULE_ID,
            severity="major",
            title="Missing inspect summary",
            description="Run is missing inspect.json, limiting architecture observability.",
            evidence=str(context.run_dir / "inspect.json"),
            suggestion="Regenerate inspect summary via pipeline finalization or `paperfig inspect`.",
        )
    ]




from __future__ import annotations

import os
import sys
import types
import unittest
from unittest.mock import patch

from paperfig.utils.paperbanana import (
    PaperBananaClient,
    PythonSDKMCPClient,
    _normalize_mcp_response,
)


class _FakeMCPClient:
    def __init__(self) -> None:
        self.calls = []

    def call_tool(self, server, tool, arguments):
        self.calls.append({"server": server, "tool": tool, "arguments": arguments})
        return {
            "svg": "<svg><text>ok</text></svg>",
            "elements": [
                {
                    "id": "e1",
                    "type": "text",
                    "label": "ok",
                    "source_spans": [
                        {
                            "section": "methodology",
                            "start": 1,
                            "end": 5,
                            "quote": "test",
                        }
                    ],
                }
            ],
        }


class PaperBananaClientTests(unittest.TestCase):
    def test_mock_mode_generates_svg_and_elements(self) -> None:
        with patch.dict(os.environ, {"PAPERFIG_MOCK_PAPERBANANA": "1"}, clear=False):
            client = PaperBananaClient()
            svg, elements = client.generate_svg({"figure_id": "fig-a", "title": "A"})

        self.assertIn("<svg", svg)
        self.assertEqual(len(elements), 1)
        self.assertEqual(elements[0]["id"], "fig-a-title")

    def test_configured_client_is_used_for_tool_call(self) -> None:
        fake = _FakeMCPClient()
        with patch.dict(os.environ, {"PAPERFIG_MOCK_PAPERBANANA": "0"}, clear=False):
            client = PaperBananaClient(mcp_client=fake, server="paperbanana-server")
            svg, elements = client.generate_svg({"figure_id": "fig-b"})

        self.assertEqual(svg, "<svg><text>ok</text></svg>")
        self.assertEqual(len(elements), 1)
        self.assertEqual(len(fake.calls), 1)
        self.assertEqual(fake.calls[0]["server"], "paperbanana-server")
        self.assertEqual(fake.calls[0]["tool"], "paperbanana.generate")

    def test_missing_configuration_raises(self) -> None:
        with patch.dict(
            os.environ,
            {
                "PAPERFIG_MOCK_PAPERBANANA": "0",
                "PAPERFIG_MCP_SERVER": "",
                "PAPERFIG_MCP_CLIENT_FACTORY": "",
                "PAPERFIG_MCP_COMMAND": "",
            },
            clear=False,
        ):
            client = PaperBananaClient()
            with self.assertRaises(RuntimeError):
                client.generate_svg({"figure_id": "fig-c"})

    def test_factory_loader_from_environment(self) -> None:
        module_name = "paperfig_test_fake_factory"
        module = types.ModuleType(module_name)

        def _factory():
            return _FakeMCPClient()

        module.build = _factory
        sys.modules[module_name] = module
        try:
            with patch.dict(
                os.environ,
                {
                    "PAPERFIG_MOCK_PAPERBANANA": "0",
                    "PAPERFIG_MCP_SERVER": "srv",
                    "PAPERFIG_MCP_CLIENT_FACTORY": f"{module_name}:build",
                    "PAPERFIG_MCP_COMMAND": "",
                },
                clear=False,
            ):
                client = PaperBananaClient()
                self.assertIsNotNone(client.mcp_client)
                svg, _ = client.generate_svg({"figure_id": "fig-d"})
                self.assertIn("<svg", svg)
        finally:
            del sys.modules[module_name]

    def test_command_configuration_creates_sdk_client(self) -> None:
        with patch.dict(
            os.environ,
            {
                "PAPERFIG_MOCK_PAPERBANANA": "0",
                "PAPERFIG_MCP_SERVER": "srv",
                "PAPERFIG_MCP_CLIENT_FACTORY": "",
                "PAPERFIG_MCP_COMMAND": "paperbanana-mcp",
                "PAPERFIG_MCP_ARGS": "--transport stdio",
            },
            clear=False,
        ):
            client = PaperBananaClient()

        self.assertIsInstance(client.mcp_client, PythonSDKMCPClient)
        self.assertEqual(client.mcp_client.command, "paperbanana-mcp")
        self.assertEqual(client.mcp_client.args, ["--transport", "stdio"])

    def test_normalize_accepts_root_svg_payload(self) -> None:
        payload = {"svg": "<svg/>", "elements": []}
        normalized = _normalize_mcp_response(payload)
        self.assertEqual(normalized["svg"], "<svg/>")

    def test_normalize_accepts_nested_result_payload(self) -> None:
        payload = {"result": {"svg": "<svg/>", "elements": []}}
        normalized = _normalize_mcp_response(payload)
        self.assertEqual(normalized["svg"], "<svg/>")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import json
import time
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, List, Optional

from paperfig.utils.types import FigureContract, FigurePlan, FlowTemplate


SCHEMA_PATH = Path("paperfig/contracts/schema/figure_contract.schema.json")


def build_figure_contract(
    run_id: str,
    plan: FigurePlan,
    template: Optional[FlowTemplate] = None,
    schema_version: str = "1.0",
) -> FigureContract:
    required_sections: List[str] = []
    if template:
        required_sections = [str(section) for section in template.required_sections]
    else:
        required_sections = [
            str(span.get("section"))
            for span in plan.source_spans
            if isinstance(span, dict) and span.get("section")
        ]

    traceability_requirements: Dict[str, Any] = {}
    if template:
        traceability_requirements = dict(template.traceability_requirements)

    invariants = ["contract_present", "traceability_present"]
    if required_sections:
        invariants.append("required_sections_present")
    if plan.source_spans:
        invariants.append("source_spans_present")

    return FigureContract(
        contract_id=f"{run_id}:{plan.figure_id}",
        schema_version=schema_version,
        run_id=run_id,
        figure_id=plan.figure_id,
        title=plan.title,
        kind=plan.kind,
        template_id=plan.template_id,
        order=plan.order,
        abstraction_level=plan.abstraction_level,
        required_sections=required_sections,
        source_spans=list(plan.source_spans),
        traceability_requirements=traceability_requirements,
        invariants=invariants,
        created_at=time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
    )


def load_contract_schema(schema_path: Path = SCHEMA_PATH) -> Dict[str, Any]:
    if not schema_path.exists():
        raise FileNotFoundError(f"Figure contract schema not found: {schema_path}")
    return json.loads(schema_path.read_text(encoding="utf-8"))


def _simple_validate(data: Dict[str, Any], schema: Dict[str, Any]) -> List[str]:
    errors: List[str] = []
    for field in schema.get("required", []):
        if field not in data:
            errors.append(f"{field}: is required")
    return errors


def _jsonschema_validate(data: Dict[str, Any], schema: Dict[str, Any]) -> List[str]:
    try:
        from jsonschema import Draft202012Validator
    except Exception:
        return _simple_validate(data, schema)

    validator = Draft202012Validator(schema)
    errors: List[str] = []
    for error in sorted(validator.iter_errors(data), key=lambda item: list(item.path)):
        field = ".".join(str(part) for part in error.path) or "<root>"
        errors.append(f"{field}: {error.message}")
    return errors


def validate_contract_data(data: Dict[str, Any], schema_path: Path = SCHEMA_PATH) -> List[str]:
    schema = load_contract_schema(schema_path=schema_path)
    return _jsonschema_validate(data, schema)


def write_contract(path: Path, contract: FigureContract) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(asdict(contract), indent=2), encoding="utf-8")


def load_contract(path: Path) -> Optional[Dict[str, Any]]:
    if not path.exists():
        return None
    try:
        return json.loads(path.read_text(encoding="utf-8"))
    except Exception:
        return None



from __future__ import annotations

import time
from pathlib import Path
from typing import Any, Dict, List

from .manifest import DocsManifest, load_manifest
from .renderer import render_hybrid_document


def _validate_required_sections(text: str, required_sections: List[str]) -> List[str]:
    missing: List[str] = []
    for section in required_sections:
        if section not in text:
            missing.append(section)
    return missing


def run_docs_regeneration(
    manifest_path: Path,
    check_only: bool,
    repo_root: Path,
) -> Dict[str, Any]:
    manifest: DocsManifest = load_manifest(manifest_path)
    doc_reports: List[Dict[str, Any]] = []
    drift_detected = False

    for entry in manifest.documents:
        doc_path = repo_root / entry.path
        report: Dict[str, Any] = {
            "path": entry.path,
            "mode": entry.mode,
            "exists": doc_path.exists(),
            "drift": False,
            "written": False,
            "missing_required_sections": [],
            "rendered_blocks": [],
            "missing_block_configs": [],
            "error": "",
        }

        if not doc_path.exists():
            report["error"] = "document_missing"
            drift_detected = True
            doc_reports.append(report)
            continue

        try:
            original = doc_path.read_text(encoding="utf-8")
            rendered = original

            if entry.mode in {"hybrid", "generated"}:
                rendered, rendered_blocks, missing_block_configs = render_hybrid_document(
                    original,
                    manifest.auto_blocks,
                    repo_root,
                )
                report["rendered_blocks"] = rendered_blocks
                report["missing_block_configs"] = missing_block_configs
                if missing_block_configs:
                    report["error"] = "missing_block_config"

            missing_sections = _validate_required_sections(rendered, entry.required_sections)
            report["missing_required_sections"] = missing_sections
            if missing_sections:
                report["error"] = "missing_required_sections"

            if rendered != original:
                report["drift"] = True
                drift_detected = True
                if not check_only:
                    doc_path.write_text(rendered, encoding="utf-8")
                    report["written"] = True
        except Exception as exc:  # pragma: no cover - defensive
            report["error"] = str(exc)
            drift_detected = True

        doc_reports.append(report)

    return {
        "checked_at": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        "manifest_path": str(manifest_path),
        "check_only": check_only,
        "drift_detected": drift_detected,
        "documents": doc_reports,
    }



from __future__ import annotations

from typing import List

from paperfig.critique.rules.base import RuleContext
from paperfig.utils.types import ArchitectureCritiqueFinding


RULE_ID = "missing_docs_drift_report"
DESCRIPTION = "Require docs drift report artifact for governance traceability."


def evaluate(context: RuleContext) -> List[ArchitectureCritiqueFinding]:
    if context.docs_drift_report is not None:
        return []
    return [
        ArchitectureCritiqueFinding(
            finding_id=RULE_ID,
            severity="minor",
            title="Missing docs drift report",
            description="No docs_drift_report.json was found for this run.",
            evidence=str(context.run_dir / "docs_drift_report.json"),
            suggestion="Enable docs check in the generation finalization stage.",
        )
    ]




from __future__ import annotations

import json
import time
from dataclasses import asdict
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Sequence, Set

from paperfig.plugins.registry import resolve_enabled_critique_plugins
from paperfig.critique.rules.base import RuleContext
from paperfig.templates.loader import load_template_catalog
from paperfig.utils.prompts import load_prompt
from paperfig.utils.types import ArchitectureCritiqueReport


SEVERITY_ORDER = {
    "info": 0,
    "minor": 1,
    "major": 2,
    "critical": 3,
}


class ArchitectureCriticAgent:
    def __init__(
        self,
        repo_root: Path = Path("."),
        template_dir: Path = Path("paperfig/templates/flows"),
        default_template_pack: str = "expanded_v1",
    ) -> None:
        self.prompt = load_prompt("critique_architecture.txt")
        self.repo_root = repo_root
        self.template_dir = template_dir
        self.default_template_pack = default_template_pack

    def available_rules(self) -> List[dict]:
        rules = resolve_enabled_critique_plugins(None)
        return [
            {
                "rule_id": plugin.descriptor.plugin_id.split(".", 1)[-1],
                "description": plugin.descriptor.description,
            }
            for plugin in rules
        ]

    def critique(
        self,
        run_dir: Path,
        block_severity: str = "critical",
        enabled_rules: Optional[Sequence[str]] = None,
    ) -> ArchitectureCritiqueReport:
        run_id = run_dir.name

        run_metadata = self._read_json(run_dir / "run.json")
        inspect_data = self._read_json(run_dir / "inspect.json")
        plan_data = self._read_json(run_dir / "plan.json")
        docs_drift_report = self._read_json(run_dir / "docs_drift_report.json")

        valid_template_ids = self._resolve_valid_template_ids(run_metadata)
        context = RuleContext(
            run_dir=run_dir,
            repo_root=self.repo_root,
            run_metadata=run_metadata if isinstance(run_metadata, dict) else {},
            inspect_data=inspect_data if isinstance(inspect_data, dict) else None,
            plan_data=plan_data if isinstance(plan_data, list) else None,
            docs_drift_report=docs_drift_report if isinstance(docs_drift_report, dict) else None,
            valid_template_ids=valid_template_ids,
        )

        findings = []
        for plugin in resolve_enabled_critique_plugins(enabled_rules):
            findings.extend(plugin.evaluator(context))

        max_seen = max((SEVERITY_ORDER.get(item.severity, 0) for item in findings), default=0)
        blocked = max_seen >= SEVERITY_ORDER.get(block_severity, SEVERITY_ORDER["critical"])

        summary = "No architecture findings."
        if findings:
            summary = f"{len(findings)} finding(s); highest severity={self._severity_name(max_seen)}"

        return ArchitectureCritiqueReport(
            run_id=run_id,
            block_severity=block_severity,
            findings=list(findings),
            blocked=blocked,
            summary=summary,
            generated_at=time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        )

    def _resolve_valid_template_ids(self, run_metadata: object) -> Set[str]:
        if not isinstance(run_metadata, dict):
            return set()

        template_pack = str(run_metadata.get("template_pack", self.default_template_pack))
        try:
            catalog = load_template_catalog(
                template_dir=self.template_dir,
                pack_id=template_pack,
                pack=template_pack,
            )
        except Exception:
            return set()
        return {template.template_id for template in catalog.templates}

    @staticmethod
    def _read_json(path: Path) -> object:
        if not path.exists():
            return None
        try:
            return json.loads(path.read_text(encoding="utf-8"))
        except Exception:
            return None

    @staticmethod
    def _severity_name(value: int) -> str:
        for name, rank in SEVERITY_ORDER.items():
            if rank == value:
                return name
        return "info"


def report_to_dict(report: ArchitectureCritiqueReport) -> Dict[str, object]:
    return asdict(report)



from __future__ import annotations

import re
from pathlib import Path
from typing import Dict, List

from .types import PaperContent, PaperSection


_HEADING_RE = re.compile(r"(?m)^(\d+\.?\s+)?([A-Za-z][A-Za-z0-9 \-]{0,80})\s*$")


def _extract_text_pdf(path: Path) -> str:
    errors: List[str] = []

    try:
        from pdfminer.high_level import extract_text  # type: ignore

        return extract_text(str(path))
    except Exception as exc:  # pragma: no cover - optional dependency
        errors.append(f"pdfminer.six failed: {exc}")

    try:
        from pypdf import PdfReader  # type: ignore

        reader = PdfReader(str(path))
        text_parts = [page.extract_text() or "" for page in reader.pages]
        return "\n".join(text_parts)
    except Exception as exc:  # pragma: no cover - optional dependency
        errors.append(f"pypdf failed: {exc}")

    raise RuntimeError(
        "PDF parsing failed. Install pdfminer.six or pypdf. Details: " + "; ".join(errors)
    )


def extract_text(path: Path) -> str:
    if path.suffix.lower() in {".md", ".markdown"}:
        return path.read_text(encoding="utf-8")
    if path.suffix.lower() == ".pdf":
        return _extract_text_pdf(path)
    raise ValueError(f"Unsupported file type: {path.suffix}")


def _find_headings(text: str) -> List[Dict[str, int | str]]:
    headings: List[Dict[str, int | str]] = []
    for match in _HEADING_RE.finditer(text):
        label = match.group(2).strip()
        # Filter out headings that are too long or likely paragraph lines.
        if len(label.split()) > 10:
            continue
        headings.append({"start": match.start(), "end": match.end(), "label": label})
    return headings


def _extract_section(text: str, name: str, keywords: List[str]) -> PaperSection:
    headings = _find_headings(text)
    for idx, heading in enumerate(headings):
        label = str(heading["label"]).lower()
        if any(keyword in label for keyword in keywords):
            start = int(heading["end"])
            end = int(headings[idx + 1]["start"]) if idx + 1 < len(headings) else len(text)
            section_text = text[start:end].strip()
            return PaperSection(name=name, text=section_text, start=start, end=end)

    # Fallback: find first keyword occurrence and capture a window
    lowered = text.lower()
    positions = [lowered.find(keyword) for keyword in keywords if lowered.find(keyword) != -1]
    if positions:
        idx = min(positions)
        start = max(0, idx - 500)
        end = min(len(text), idx + 2000)
        return PaperSection(name=name, text=text[start:end].strip(), start=start, end=end)

    return PaperSection(name=name, text="", start=0, end=0)


def extract_sections(text: str) -> Dict[str, PaperSection]:
    methodology = _extract_section(
        text,
        name="methodology",
        keywords=["method", "methods", "methodology", "approach"],
    )
    system = _extract_section(
        text,
        name="system",
        keywords=["system", "architecture", "model", "pipeline"],
    )
    results = _extract_section(
        text,
        name="results",
        keywords=["results", "experiments", "evaluation"],
    )
    return {
        "methodology": methodology,
        "system": system,
        "results": results,
    }


def parse_paper(path: Path) -> PaperContent:
    text = extract_text(path)
    sections = extract_sections(text)
    return PaperContent(source_path=str(path), full_text=text, sections=sections)



from __future__ import annotations

import json
import time
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, List

from paperfig.contracts import load_contract, validate_contract_data
from paperfig.utils.types import InspectHtmlManifest


HTML_TITLE = "paperfig inspector"


def build_html_inspector(run_dir: Path, run_root: Path) -> InspectHtmlManifest:
    run_id = run_dir.name
    inspect_data = _read_json(run_dir / "inspect.json")
    architecture = _read_json(run_dir / "architecture_critique.json")
    repro = _read_json(run_dir / "repro_audit.json")
    docs_drift = _read_json(run_dir / "docs_drift_report.json")

    figures = _collect_figures(run_dir)
    diffs = _collect_diffs(run_root, run_id)

    payload = {
        "run_id": run_id,
        "inspect": inspect_data,
        "architecture_critique": architecture,
        "repro_audit": repro,
        "docs_drift": docs_drift,
        "figures": figures,
        "diffs": diffs,
    }

    inspect_dir = run_dir / "inspect"
    inspect_dir.mkdir(parents=True, exist_ok=True)
    html_path = inspect_dir / "index.html"
    html_path.write_text(_render_html(payload), encoding="utf-8")

    manifest = InspectHtmlManifest(
        run_id=run_id,
        generated_at=time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        html_path=str(html_path),
        artifacts=[
            "inspect/index.html",
            "inspect/manifest.json",
            "inspect.json",
            "architecture_critique.json",
            "repro_audit.json",
            "docs_drift_report.json",
            "figures/*/final/figure.svg",
            "figures/*/contract.json",
        ],
        summary={
            "figure_count": len(figures),
            "diff_count": len(diffs),
        },
    )
    manifest_path = inspect_dir / "manifest.json"
    manifest_path.write_text(json.dumps(asdict(manifest), indent=2), encoding="utf-8")
    return manifest


def _collect_figures(run_dir: Path) -> List[Dict[str, Any]]:
    figures: List[Dict[str, Any]] = []
    figures_dir = run_dir / "figures"
    if not figures_dir.exists():
        return figures

    for figure_dir in sorted([path for path in figures_dir.iterdir() if path.is_dir()]):
        final_dir = figure_dir / "final"
        svg_path = final_dir / "figure.svg"
        traceability_path = final_dir / "traceability.json"
        contract_path = figure_dir / "contract.json"

        svg_text = svg_path.read_text(encoding="utf-8") if svg_path.exists() else None
        traceability = _read_json(traceability_path)
        contract = load_contract(contract_path) if contract_path.exists() else None
        contract_errors: List[str] = []
        if isinstance(contract, dict):
            contract_errors = validate_contract_data(contract)

        critique = _latest_critique(figure_dir)

        figures.append(
            {
                "figure_id": figure_dir.name,
                "svg": svg_text,
                "traceability": traceability,
                "contract": contract,
                "contract_errors": contract_errors,
                "critique": critique,
                "final_svg_path": str(svg_path) if svg_path.exists() else None,
            }
        )

    return figures


def _latest_critique(figure_dir: Path) -> Dict[str, Any] | None:
    iter_dirs = sorted(
        [path for path in figure_dir.glob("iter_*") if path.is_dir()],
        key=lambda p: int(p.name.split("_")[1]) if "_" in p.name and p.name.split("_")[1].isdigit() else 0,
    )
    for iter_dir in reversed(iter_dirs):
        critique_path = iter_dir / "critique.json"
        if critique_path.exists():
            return _read_json(critique_path)
    return None


def _collect_diffs(run_root: Path, run_id: str) -> List[Dict[str, Any]]:
    diffs_root = run_root / "diffs"
    if not diffs_root.exists():
        return []
    reports: List[Dict[str, Any]] = []
    for diff_path in diffs_root.glob("**/diff.json"):
        data = _read_json(diff_path)
        if not isinstance(data, dict):
            continue
        if run_id in {data.get("run_id_1"), data.get("run_id_2")}:
            data = dict(data)
            data["diff_path"] = str(diff_path)
            reports.append(data)
    return reports


def _read_json(path: Path) -> Dict[str, Any] | None:
    if not path.exists():
        return None
    try:
        return json.loads(path.read_text(encoding="utf-8"))
    except Exception:
        return None


def _render_html(payload: Dict[str, Any]) -> str:
    data_json = json.dumps(payload, indent=2).replace("</", "<\\/")

    template = """<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>__TITLE__</title>
  <style>
    body { font-family: ui-sans-serif, system-ui, -apple-system, Segoe UI, Roboto, Helvetica, Arial; margin: 24px; color: #1f2933; }
    h1 { font-size: 24px; margin-bottom: 8px; }
    h2 { font-size: 18px; margin-top: 24px; }
    .meta { color: #52606d; margin-bottom: 16px; }
    .card { border: 1px solid #e0e4e8; border-radius: 8px; padding: 16px; margin-bottom: 16px; background: #fafbfc; }
    .grid { display: grid; grid-template-columns: repeat(auto-fit, minmax(280px, 1fr)); gap: 16px; }
    .svg-box { border: 1px dashed #9fb3c8; padding: 8px; background: #fff; overflow: auto; }
    pre { white-space: pre-wrap; background: #f5f7fa; padding: 12px; border-radius: 6px; }
    .badge { display: inline-block; padding: 2px 8px; border-radius: 12px; font-size: 12px; background: #d9e2ec; margin-right: 6px; }
    .warn { background: #ffe3c5; }
    .ok { background: #c6f6d5; }
  </style>
</head>
<body>
  <h1>__TITLE__</h1>
  <div class="meta">Run: <span id="run-id"></span></div>

  <div class="card">
    <h2>Summary</h2>
    <div id="summary"></div>
  </div>

  <div class="card">
    <h2>Architecture Critique</h2>
    <pre id="architecture"></pre>
  </div>

  <div class="card">
    <h2>Reproducibility Audit</h2>
    <pre id="repro"></pre>
  </div>

  <div class="card">
    <h2>Docs Drift Report</h2>
    <pre id="docs-drift"></pre>
  </div>

  <div class="card">
    <h2>Figures</h2>
    <div id="figures" class="grid"></div>
  </div>

  <div class="card">
    <h2>Diff Reports</h2>
    <pre id="diffs"></pre>
  </div>

  <script id="inspect-data" type="application/json">__DATA__</script>
  <script>
    const data = JSON.parse(document.getElementById('inspect-data').textContent);
    document.getElementById('run-id').textContent = data.run_id || 'unknown';

    const inspect = data.inspect || {};
    const aggregate = (inspect.aggregate || {});
    document.getElementById('summary').innerHTML = `
      <div class="badge">Figures: ${aggregate.accepted_count || 0}/${aggregate.total_figures || 0} accepted</div>
      <div class="badge">Avg score: ${aggregate.avg_final_score ?? 'n/a'}</div>
      <div class="badge">Avg traceability: ${aggregate.avg_traceability_coverage ?? 'n/a'}</div>
    `;

    document.getElementById('architecture').textContent = JSON.stringify(data.architecture_critique || {}, null, 2);
    document.getElementById('repro').textContent = JSON.stringify(data.repro_audit || {}, null, 2);
    document.getElementById('docs-drift').textContent = JSON.stringify(data.docs_drift || {}, null, 2);
    document.getElementById('diffs').textContent = JSON.stringify(data.diffs || [], null, 2);

    const figuresEl = document.getElementById('figures');
    (data.figures || []).forEach(fig => {
      const card = document.createElement('div');
      card.className = 'card';
      const contractErrors = fig.contract_errors || [];
      const contractBadge = contractErrors.length ? '<span class="badge warn">contract issues</span>' : '<span class="badge ok">contract ok</span>';
      card.innerHTML = `
        <h3>${fig.figure_id}</h3>
        ${contractBadge}
        <div class="svg-box">${fig.svg || '<em>SVG not found</em>'}</div>
        <h4>Critique</h4>
        <pre>${JSON.stringify(fig.critique || {}, null, 2)}</pre>
        <h4>Traceability</h4>
        <pre>${JSON.stringify(fig.traceability || {}, null, 2)}</pre>
        <h4>Contract</h4>
        <pre>${JSON.stringify(fig.contract || {}, null, 2)}</pre>
      `;
      figuresEl.appendChild(card);
    });
  </script>
</body>
</html>
"""

    return template.replace("__TITLE__", HTML_TITLE).replace("__DATA__", data_json)



from __future__ import annotations

import shutil
from pathlib import Path


def export_svg(src: Path, dest: Path) -> None:
    dest.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy2(src, dest)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List


@dataclass
class LabPolicy:
    allowed_prefixes: List[str] = field(default_factory=list)
    blocked_patterns: List[str] = field(default_factory=list)
    max_runtime_seconds: int = 1200
    max_parallel_experiments: int = 1


@dataclass
class LabExperimentSpec:
    experiment_id: str
    topic: str
    source_run_id: str
    hypothesis: str
    command: str
    status: str = "proposed"
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class LabExperimentResult:
    experiment_id: str
    status: str
    return_code: int
    started_at: str
    finished_at: str
    stdout: str
    stderr: str
    policy_violation: str = ""



from __future__ import annotations

import json
import tempfile
import types
import unittest
from pathlib import Path
from unittest.mock import patch

try:
    from typer.testing import CliRunner
    from paperfig.cli import app
except ModuleNotFoundError:  # pragma: no cover - environment fallback
    CliRunner = None  # type: ignore[assignment]
    app = None  # type: ignore[assignment]


@unittest.skipIf(CliRunner is None or app is None, "typer is not installed in this environment")
class CliExtensionsTests(unittest.TestCase):
    def setUp(self) -> None:
        self.runner = CliRunner()

    def test_critique_architecture_list_rules(self) -> None:
        result = self.runner.invoke(app, ["critique-architecture", "--list-rules"])
        self.assertEqual(result.exit_code, 0, msg=result.stdout)
        self.assertIn("missing_flow_docs", result.stdout)
        self.assertIn("traceability_gap", result.stdout)
        self.assertIn("invalid_template_reference", result.stdout)

    def test_critique_architecture_enable_filter(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            run_root = Path(tmpdir) / "runs"
            run_dir = run_root / "run-filter"
            run_dir.mkdir(parents=True, exist_ok=True)
            (run_dir / "inspect.json").write_text(
                json.dumps({"aggregate": {"failed_count": 0, "avg_traceability_coverage": 0.4}}),
                encoding="utf-8",
            )

            result = self.runner.invoke(
                app,
                [
                    "critique-architecture",
                    "run-filter",
                    "--run-root",
                    str(run_root),
                    "--enable",
                    "traceability_gap",
                    "--as-json",
                ],
            )
            self.assertEqual(result.exit_code, 0, msg=result.stdout)
            payload = json.loads(result.stdout)
            ids = {item["finding_id"] for item in payload.get("findings", [])}
            self.assertEqual(ids, {"traceability_gap"})

    def test_doctor_missing_cairosvg_shows_fix_hint(self) -> None:
        def _fake_import(name: str, *args, **kwargs):  # type: ignore[no-untyped-def]
            del args, kwargs
            if name == "cairosvg":
                raise ImportError("missing")
            return types.SimpleNamespace(__name__=name)

        with patch("paperfig.cli.importlib.import_module", side_effect=_fake_import):
            result = self.runner.invoke(app, ["doctor"])

        self.assertEqual(result.exit_code, 0, msg=result.stdout)
        self.assertIn("paperfig doctor --fix png", result.stdout)

    def test_plugins_list(self) -> None:
        result = self.runner.invoke(app, ["plugins", "list"])
        self.assertEqual(result.exit_code, 0, msg=result.stdout)
        self.assertIn("critique_rule", result.stdout)
        self.assertIn("repro_check", result.stdout)

    def test_plugins_validate(self) -> None:
        result = self.runner.invoke(app, ["plugins", "validate"])
        self.assertEqual(result.exit_code, 0, msg=result.stdout)
        self.assertIn("Plugin registry is valid", result.stdout)


if __name__ == "__main__":
    unittest.main()



from .base import PluginBase, CritiqueRulePlugin, ReproCheckPlugin
from .registry import list_plugins, validate_plugins, resolve_enabled_critique_plugins, get_repro_plugins

__all__ = [
    "PluginBase",
    "CritiqueRulePlugin",
    "ReproCheckPlugin",
    "list_plugins",
    "validate_plugins",
    "resolve_enabled_critique_plugins",
    "get_repro_plugins",
]



from __future__ import annotations

from typing import Dict, Iterable, List, Optional

from . import (
    empty_plan,
    failed_figures,
    invalid_template_reference,
    missing_docs_drift_report,
    missing_flow_docs,
    missing_inspect,
    missing_plan,
    traceability_gap,
)
from .base import ArchitectureRule


def get_rule_registry() -> Dict[str, ArchitectureRule]:
    rules = [
        ArchitectureRule(
            rule_id=missing_inspect.RULE_ID,
            description=missing_inspect.DESCRIPTION,
            evaluator=missing_inspect.evaluate,
        ),
        ArchitectureRule(
            rule_id=failed_figures.RULE_ID,
            description=failed_figures.DESCRIPTION,
            evaluator=failed_figures.evaluate,
        ),
        ArchitectureRule(
            rule_id=traceability_gap.RULE_ID,
            description=traceability_gap.DESCRIPTION,
            evaluator=traceability_gap.evaluate,
        ),
        ArchitectureRule(
            rule_id=missing_plan.RULE_ID,
            description=missing_plan.DESCRIPTION,
            evaluator=missing_plan.evaluate,
        ),
        ArchitectureRule(
            rule_id=empty_plan.RULE_ID,
            description=empty_plan.DESCRIPTION,
            evaluator=empty_plan.evaluate,
        ),
        ArchitectureRule(
            rule_id=missing_docs_drift_report.RULE_ID,
            description=missing_docs_drift_report.DESCRIPTION,
            evaluator=missing_docs_drift_report.evaluate,
        ),
        ArchitectureRule(
            rule_id=missing_flow_docs.RULE_ID,
            description=missing_flow_docs.DESCRIPTION,
            evaluator=missing_flow_docs.evaluate,
        ),
        ArchitectureRule(
            rule_id=invalid_template_reference.RULE_ID,
            description=invalid_template_reference.DESCRIPTION,
            evaluator=invalid_template_reference.evaluate,
        ),
    ]
    return {rule.rule_id: rule for rule in rules}


def list_rule_descriptors() -> List[dict]:
    registry = get_rule_registry()
    return [
        {"rule_id": rule_id, "description": rule.description}
        for rule_id, rule in sorted(registry.items())
    ]


def resolve_enabled_rules(enable: Optional[Iterable[str]]) -> List[ArchitectureRule]:
    registry = get_rule_registry()
    if not enable:
        return [registry[key] for key in sorted(registry.keys())]

    selected: List[ArchitectureRule] = []
    for rule_id in enable:
        if rule_id not in registry:
            available = ", ".join(sorted(registry.keys()))
            raise ValueError(f"Unknown architecture rule '{rule_id}'. Available: {available}")
        selected.append(registry[rule_id])
    return selected



from __future__ import annotations

import re
from pathlib import Path
from typing import Dict, List, Tuple

from paperfig.command_catalog import get_command_catalog
from paperfig.templates.loader import load_template_catalog


AUTO_BLOCK_RE = re.compile(
    r"<!--\s*AUTO-GEN:START\s+(?P<block_id>[A-Za-z0-9_\-]+)\s*-->"
    r"(?P<body>.*?)"
    r"<!--\s*AUTO-GEN:END\s+(?P=block_id)\s*-->",
    re.DOTALL,
)


def render_auto_block(block_id: str, block_config: Dict[str, object], repo_root: Path) -> str:
    block_type = str(block_config.get("type", ""))

    if block_type == "cli_commands":
        commands = get_command_catalog()
        return "\n" + "\n".join(f"- `paperfig {command}`" for command in commands) + "\n"

    if block_type == "flow_template_catalog":
        template_dir = repo_root / str(block_config.get("template_dir", "paperfig/templates/flows"))
        pack_id = str(block_config.get("pack_id", "expanded_v1"))
        catalog = load_template_catalog(template_dir=template_dir, pack_id=pack_id)
        lines = [f"- `{tmpl.template_id}` ({tmpl.kind})" for tmpl in catalog.templates]
        return "\n" + "\n".join(lines) + "\n"

    if block_type == "static":
        content = str(block_config.get("content", ""))
        return "\n" + content.strip("\n") + "\n"

    raise RuntimeError(f"Unknown auto block type for '{block_id}': {block_type}")


def render_hybrid_document(
    text: str,
    auto_blocks: Dict[str, Dict[str, object]],
    repo_root: Path,
) -> Tuple[str, List[str], List[str]]:
    rendered_blocks: List[str] = []
    missing_block_configs: List[str] = []

    def _replace(match: re.Match[str]) -> str:
        block_id = match.group("block_id")
        block_config = auto_blocks.get(block_id)
        if not isinstance(block_config, dict):
            missing_block_configs.append(block_id)
            return match.group(0)

        rendered = render_auto_block(block_id, block_config, repo_root)
        rendered_blocks.append(block_id)
        return f"<!-- AUTO-GEN:START {block_id} -->{rendered}<!-- AUTO-GEN:END {block_id} -->"

    rendered_text = AUTO_BLOCK_RE.sub(_replace, text)
    return rendered_text, rendered_blocks, missing_block_configs



from __future__ import annotations

from typing import List

from paperfig.critique.rules.base import RuleContext
from paperfig.utils.types import ArchitectureCritiqueFinding


RULE_ID = "missing_plan"
DESCRIPTION = "Require run plan artifact."


def evaluate(context: RuleContext) -> List[ArchitectureCritiqueFinding]:
    if context.plan_data is not None:
        return []
    return [
        ArchitectureCritiqueFinding(
            finding_id=RULE_ID,
            severity="critical",
            title="Missing plan artifact",
            description="Run is missing plan.json.",
            evidence=str(context.run_dir / "plan.json"),
            suggestion="Investigate planner stage and regenerate run artifacts.",
        )
    ]




from .drift import run_docs_regeneration
from .manifest import DocsManifest, DocManifestEntry, load_manifest

__all__ = [
    "run_docs_regeneration",
    "DocsManifest",
    "DocManifestEntry",
    "load_manifest",
]



from __future__ import annotations

from typing import Dict

from paperfig.lab.types import LabExperimentResult, LabExperimentSpec


def review_experiment(spec: LabExperimentSpec, result: LabExperimentResult) -> Dict[str, object]:
    recommendation = "revise"
    rationale = "Execution failed or produced policy violations."

    if result.status == "completed" and result.return_code == 0 and not result.policy_violation:
        recommendation = "promote"
        rationale = "Execution completed successfully and respected sandbox policy."

    return {
        "experiment_id": spec.experiment_id,
        "recommendation": recommendation,
        "rationale": rationale,
        "status": result.status,
        "return_code": result.return_code,
        "policy_violation": result.policy_violation,
    }



from __future__ import annotations

import platform
import sys
import time
from dataclasses import asdict
from pathlib import Path
from typing import Dict, List

from paperfig.audits.repro_checks import get_repro_check_registry
from paperfig.utils.types import ReproAuditReport


def run_reproducibility_audit(
    run_dir: Path,
    mode: str = "soft",
    expected_config_hash: str | None = None,
) -> ReproAuditReport:
    run_id = run_dir.name
    checks = []
    registry = get_repro_check_registry()
    for check in registry.values():
        checks.append(check.evaluator(run_dir, expected_config_hash))

    required_failed = [check for check in checks if check.required and not check.passed]
    passed = len(required_failed) == 0

    summary = "Reproducibility checks passed."
    if not passed:
        summary = f"{len(required_failed)} required reproducibility check(s) failed."

    return ReproAuditReport(
        run_id=run_id,
        mode=mode,
        checks=checks,
        passed=passed,
        summary=summary,
        generated_at=time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        environment={
            "python_version": sys.version,
            "platform": platform.platform(),
        },
    )


def report_to_dict(report: ReproAuditReport) -> Dict[str, object]:
    return asdict(report)



from __future__ import annotations

import re
import unittest
from pathlib import Path


class RenameAlignmentTests(unittest.TestCase):
    def test_pyproject_package_name_and_console_script(self) -> None:
        content = Path("pyproject.toml").read_text(encoding="utf-8")
        name_match = re.search(r'name\s*=\s*"([^"]+)"', content)
        self.assertIsNotNone(name_match, msg="Missing [project].name in pyproject.toml")
        self.assertEqual(name_match.group(1), "paperfigg")

        script_match = re.search(r'paperfig\s*=\s*"([^"]+)"', content)
        self.assertIsNotNone(script_match, msg="Missing console script for paperfig")
        self.assertEqual(script_match.group(1), "paperfig.cli:app")

    def test_readme_has_guardrail_and_correct_installs(self) -> None:
        readme = Path("README.md").read_text(encoding="utf-8")
        self.assertRegex(
            readme,
            r"`paperfig` on PyPI is a different project; install `paperfigg`",
        )
        self.assertNotIn('pip install "paperfig"', readme)
        self.assertIn('pip install "paperfigg', readme)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from pathlib import Path


def export_latex(figure_id: str, svg_filename: str, caption: str, output_path: Path) -> None:
    output_path.parent.mkdir(parents=True, exist_ok=True)
    snippet = (
        "\\begin{figure}[t]\n"
        "  \\centering\n"
        f"  \\includegraphics[width=\\linewidth]{{{svg_filename}}}\n"
        f"  \\caption{{{caption}}}\n"
        f"  \\label{{fig:{figure_id}}}\n"
        "\\end{figure}\n"
    )
    output_path.write_text(snippet, encoding="utf-8")



from __future__ import annotations

import shlex
from pathlib import Path
from typing import Tuple

from paperfig.lab.types import LabPolicy
from paperfig.utils.structured_data import load_structured_file


def load_policy(path: Path) -> LabPolicy:
    data = load_structured_file(path)
    if not isinstance(data, dict):
        raise RuntimeError(f"Policy file {path} must contain a mapping/object.")

    return LabPolicy(
        allowed_prefixes=[str(item) for item in data.get("allowed_prefixes", [])],
        blocked_patterns=[str(item) for item in data.get("blocked_patterns", [])],
        max_runtime_seconds=int(data.get("max_runtime_seconds", 1200)),
        max_parallel_experiments=int(data.get("max_parallel_experiments", 1)),
    )


def is_command_allowed(command: str, policy: LabPolicy) -> Tuple[bool, str]:
    lowered = command.lower()
    for pattern in policy.blocked_patterns:
        if pattern.lower() in lowered:
            return False, f"Command blocked by pattern: {pattern}"

    tokens = shlex.split(command)
    if not tokens:
        return False, "Command is empty"

    prefix = tokens[0]
    if policy.allowed_prefixes and prefix not in policy.allowed_prefixes:
        if prefix == "python" and "python3" in policy.allowed_prefixes:
            return True, "allowed (python alias for python3)"
        return False, f"Command prefix '{prefix}' is not allowed"

    return True, "allowed"
